"""Representing the artifacts of a project."""

from itertools import count
import logging

from typing import TYPE_CHECKING, TypeVar, Type, Optional

if TYPE_CHECKING:
    from ..project import Project
    from ..knowledge_plugins import FunctionManager
    from ..knowledge_plugins import VariableManager
    from ..knowledge_plugins import KeyDefinitionManager
    from ..knowledge_plugins import CFGManager
    from ..knowledge_plugins import StructuredCodeManager
    from ..knowledge_plugins import TypesStore

from ..knowledge_plugins.plugin import default_plugins, KnowledgeBasePlugin

l = logging.getLogger(name=__name__)


kb_ctr = count(0, 1)


class KnowledgeBase:
    """Represents a "model" of knowledge about an artifact.

    Contains things like a CFG, data references, etc.
    """
    functions: 'FunctionManager'
    variables: 'VariableManager'
    structured_code: 'StructuredCodeManager'
    defs: 'KeyDefinitionManager'
    cfgs: 'CFGManager'
    _project: 'Project'
    types: 'TypesStore'

    def __init__(self, project, obj=None, name=None):
        if obj is not None:
            l.warning("The obj parameter in KnowledgeBase.__init__() has been deprecated.")
        object.__setattr__(self, '_project', project)
        object.__setattr__(self, '_plugins', {})

        self.name = name if name else ("kb_%d" % next(kb_ctr))

    @property
    def callgraph(self):
        return self.functions.callgraph

    @property
    def unresolved_indirect_jumps(self):
        return self.indirect_jumps.unresolved

    @property
    def resolved_indirect_jumps(self):
        return self.indirect_jumps.resolved

    def __setstate__(self, state):
        object.__setattr__(self, '_project', state['project'])
        object.__setattr__(self, '_plugins', state['plugins'])

    def __getstate__(self):
        s = {
            'project': self._project,
            'plugins': self._plugins,
        }
        return s

    def __dir__(self):
        x = list(super(KnowledgeBase, self).__dir__())
        x.extend(default_plugins.keys())
        return x

    #
    # Plugin accessor
    #

    def __contains__(self, plugin_name):
        return plugin_name in self._plugins

    def __getattr__(self, v):
        try:
            return self.get_plugin(v)
        except KeyError:
            raise AttributeError(v)

    def __setattr__(self, k, v):
        self.register_plugin(k, v)

    #
    # Plugins
    #

    def has_plugin(self, name):
        return name in self._plugins

    def get_plugin(self, name):
        if name not in self._plugins:
            p = default_plugins[name](self)
            self.register_plugin(name, p)
            return p
        return self._plugins[name]

    def register_plugin(self, name, plugin):
        self._plugins[name] = plugin
        return plugin

    def release_plugin(self, name):
        if name in self._plugins:
            del self._plugins[name]

    K = TypeVar("K", bound=KnowledgeBasePlugin)
    def get_knowledge(self, requested_plugin_cls: Type[K]) -> Optional[K]:
        """
        Type inference safe method to request a knowledge base plugin
        Explicitly passing the type of the requested plugin achieves two things:
        1. Every location using this plugin can be easily found with an IDE by searching explicit references to the type
        2. Basic type inference can deduce the result type and properly type check usages of it

        If there isn't already an instance of this class None will be returned to make it clear to the caller that there
        is no existing knowledge of this type yet. The code that initially creates this knowledge should use the
        `register_plugin` method to register the initial knowledge state
        :param requested_plugin_cls:
        :return: Instance of the requested plugin class or null if it is not a known plugin
        """
        # Get first plugin of this type already registered, or default to None
        return next(
            filter(lambda registered_plugin: type(registered_plugin) == requested_plugin_cls, self._plugins.values()),
            None)

    def request_knowledge(self, requested_plugin_cls: Type[K]) -> K:
        existing = self.get_knowledge(requested_plugin_cls)
        if existing is not None:
            return existing
        else:
            p = requested_plugin_cls(self)
            self.register_plugin(requested_plugin_cls.__name__, p)
            return p



import angr

import logging
l = logging.getLogger(name=__name__)


class atoi(angr.SimProcedure):
    #pylint:disable=arguments-differ
    def run(self, s):
        strtol = angr.SIM_PROCEDURES['libc']['strtol']
        val = strtol.strtol_inner(s, self.state, self.state.memory, 10, True)[1]
        val = val[self.arch.sizeof['int'] - 1:0]
        return val



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: protos/variables.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x16protos/variables.proto\x12\x0b\x61ngr.protos\"\x90\x01\n\x0cVariableBase\x12\r\n\x05ident\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x13\n\x06region\x18\x03 \x01(\x04H\x00\x88\x01\x01\x12\x15\n\x08\x63\x61tegory\x18\x04 \x01(\tH\x01\x88\x01\x01\x12\x0f\n\x07renamed\x18\x05 \x01(\x08\x12\x0e\n\x06is_phi\x18\x06 \x01(\x08\x42\t\n\x07_regionB\x0b\n\t_category\"L\n\x11TemporaryVariable\x12\'\n\x04\x62\x61se\x18\x01 \x01(\x0b\x32\x19.angr.protos.VariableBase\x12\x0e\n\x06tmp_id\x18\x02 \x01(\r\"V\n\x10RegisterVariable\x12\'\n\x04\x62\x61se\x18\x01 \x01(\x0b\x32\x19.angr.protos.VariableBase\x12\x0b\n\x03reg\x18\x02 \x01(\r\x12\x0c\n\x04size\x18\x03 \x01(\r\"U\n\x0eMemoryVariable\x12\'\n\x04\x62\x61se\x18\x01 \x01(\x0b\x32\x19.angr.protos.VariableBase\x12\x0c\n\x04\x61\x64\x64r\x18\x02 \x01(\x04\x12\x0c\n\x04size\x18\x03 \x01(\r\"u\n\rStackVariable\x12\'\n\x04\x62\x61se\x18\x01 \x01(\x0b\x32\x19.angr.protos.VariableBase\x12\x0c\n\x04\x61\x64\x64r\x18\x02 \x01(\x04\x12\x0c\n\x04size\x18\x03 \x01(\r\x12\x0f\n\x07sp_base\x18\x04 \x01(\x08\x12\x0e\n\x06offset\x18\x05 \x01(\x05\"\x9c\x02\n\x0eVariableAccess\x12\r\n\x05ident\x18\x01 \x01(\t\x12\x12\n\nblock_addr\x18\x02 \x01(\x04\x12\x10\n\x08stmt_idx\x18\x03 \x01(\x05\x12\x10\n\x08ins_addr\x18\x04 \x01(\x04\x12\x13\n\x06offset\x18\x05 \x01(\x03H\x00\x88\x01\x01\x12\x43\n\x0b\x61\x63\x63\x65ss_type\x18\x06 \x01(\x0e\x32..angr.protos.VariableAccess.VariableAccessSort\x12\x16\n\tatom_hash\x18\x07 \x01(\rH\x01\x88\x01\x01\"8\n\x12VariableAccessSort\x12\t\n\x05WRITE\x10\x00\x12\x08\n\x04READ\x10\x01\x12\r\n\tREFERENCE\x10\x02\x42\t\n\x07_offsetB\x0c\n\n_atom_hash\"/\n\x0cVariableType\x12\r\n\x05ident\x18\x01 \x01(\t\x12\x10\n\x08var_type\x18\x02 \x01(\t\";\n\x0bVar2Unified\x12\x11\n\tvar_ident\x18\x01 \x01(\t\x12\x19\n\x11unified_var_ident\x18\x02 \x01(\t\"/\n\x07Phi2Var\x12\x11\n\tphi_ident\x18\x01 \x01(\t\x12\x11\n\tvar_ident\x18\x02 \x01(\t\"\xe6\x04\n\x17VariableManagerInternal\x12\x30\n\x08tempvars\x18\x01 \x03(\x0b\x32\x1e.angr.protos.TemporaryVariable\x12.\n\x07regvars\x18\x02 \x03(\x0b\x32\x1d.angr.protos.RegisterVariable\x12,\n\x07memvars\x18\x03 \x03(\x0b\x32\x1b.angr.protos.MemoryVariable\x12-\n\tstackvars\x18\x04 \x03(\x0b\x32\x1a.angr.protos.StackVariable\x12-\n\x08\x61\x63\x63\x65sses\x18\x05 \x03(\x0b\x32\x1b.angr.protos.VariableAccess\x12\x38\n\x10unified_tempvars\x18\x06 \x03(\x0b\x32\x1e.angr.protos.TemporaryVariable\x12\x36\n\x0funified_regvars\x18\x07 \x03(\x0b\x32\x1d.angr.protos.RegisterVariable\x12\x34\n\x0funified_memvars\x18\x08 \x03(\x0b\x32\x1b.angr.protos.MemoryVariable\x12\x35\n\x11unified_stackvars\x18\t \x03(\x0b\x32\x1a.angr.protos.StackVariable\x12-\n\x0bvar2unified\x18\n \x03(\x0b\x32\x18.angr.protos.Var2Unified\x12(\n\x05types\x18\x0b \x03(\x0b\x32\x19.angr.protos.VariableType\x12%\n\x07phi2var\x18\x0c \x03(\x0b\x32\x14.angr.protos.Phi2Varb\x06proto3')



_VARIABLEBASE = DESCRIPTOR.message_types_by_name['VariableBase']
_TEMPORARYVARIABLE = DESCRIPTOR.message_types_by_name['TemporaryVariable']
_REGISTERVARIABLE = DESCRIPTOR.message_types_by_name['RegisterVariable']
_MEMORYVARIABLE = DESCRIPTOR.message_types_by_name['MemoryVariable']
_STACKVARIABLE = DESCRIPTOR.message_types_by_name['StackVariable']
_VARIABLEACCESS = DESCRIPTOR.message_types_by_name['VariableAccess']
_VARIABLETYPE = DESCRIPTOR.message_types_by_name['VariableType']
_VAR2UNIFIED = DESCRIPTOR.message_types_by_name['Var2Unified']
_PHI2VAR = DESCRIPTOR.message_types_by_name['Phi2Var']
_VARIABLEMANAGERINTERNAL = DESCRIPTOR.message_types_by_name['VariableManagerInternal']
_VARIABLEACCESS_VARIABLEACCESSSORT = _VARIABLEACCESS.enum_types_by_name['VariableAccessSort']
VariableBase = _reflection.GeneratedProtocolMessageType('VariableBase', (_message.Message,), {
  'DESCRIPTOR' : _VARIABLEBASE,
  '__module__' : 'protos.variables_pb2'
  # @@protoc_insertion_point(class_scope:angr.protos.VariableBase)
  })
_sym_db.RegisterMessage(VariableBase)

TemporaryVariable = _reflection.GeneratedProtocolMessageType('TemporaryVariable', (_message.Message,), {
  'DESCRIPTOR' : _TEMPORARYVARIABLE,
  '__module__' : 'protos.variables_pb2'
  # @@protoc_insertion_point(class_scope:angr.protos.TemporaryVariable)
  })
_sym_db.RegisterMessage(TemporaryVariable)

RegisterVariable = _reflection.GeneratedProtocolMessageType('RegisterVariable', (_message.Message,), {
  'DESCRIPTOR' : _REGISTERVARIABLE,
  '__module__' : 'protos.variables_pb2'
  # @@protoc_insertion_point(class_scope:angr.protos.RegisterVariable)
  })
_sym_db.RegisterMessage(RegisterVariable)

MemoryVariable = _reflection.GeneratedProtocolMessageType('MemoryVariable', (_message.Message,), {
  'DESCRIPTOR' : _MEMORYVARIABLE,
  '__module__' : 'protos.variables_pb2'
  # @@protoc_insertion_point(class_scope:angr.protos.MemoryVariable)
  })
_sym_db.RegisterMessage(MemoryVariable)

StackVariable = _reflection.GeneratedProtocolMessageType('StackVariable', (_message.Message,), {
  'DESCRIPTOR' : _STACKVARIABLE,
  '__module__' : 'protos.variables_pb2'
  # @@protoc_insertion_point(class_scope:angr.protos.StackVariable)
  })
_sym_db.RegisterMessage(StackVariable)

VariableAccess = _reflection.GeneratedProtocolMessageType('VariableAccess', (_message.Message,), {
  'DESCRIPTOR' : _VARIABLEACCESS,
  '__module__' : 'protos.variables_pb2'
  # @@protoc_insertion_point(class_scope:angr.protos.VariableAccess)
  })
_sym_db.RegisterMessage(VariableAccess)

VariableType = _reflection.GeneratedProtocolMessageType('VariableType', (_message.Message,), {
  'DESCRIPTOR' : _VARIABLETYPE,
  '__module__' : 'protos.variables_pb2'
  # @@protoc_insertion_point(class_scope:angr.protos.VariableType)
  })
_sym_db.RegisterMessage(VariableType)

Var2Unified = _reflection.GeneratedProtocolMessageType('Var2Unified', (_message.Message,), {
  'DESCRIPTOR' : _VAR2UNIFIED,
  '__module__' : 'protos.variables_pb2'
  # @@protoc_insertion_point(class_scope:angr.protos.Var2Unified)
  })
_sym_db.RegisterMessage(Var2Unified)

Phi2Var = _reflection.GeneratedProtocolMessageType('Phi2Var', (_message.Message,), {
  'DESCRIPTOR' : _PHI2VAR,
  '__module__' : 'protos.variables_pb2'
  # @@protoc_insertion_point(class_scope:angr.protos.Phi2Var)
  })
_sym_db.RegisterMessage(Phi2Var)

VariableManagerInternal = _reflection.GeneratedProtocolMessageType('VariableManagerInternal', (_message.Message,), {
  'DESCRIPTOR' : _VARIABLEMANAGERINTERNAL,
  '__module__' : 'protos.variables_pb2'
  # @@protoc_insertion_point(class_scope:angr.protos.VariableManagerInternal)
  })
_sym_db.RegisterMessage(VariableManagerInternal)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _VARIABLEBASE._serialized_start=40
  _VARIABLEBASE._serialized_end=184
  _TEMPORARYVARIABLE._serialized_start=186
  _TEMPORARYVARIABLE._serialized_end=262
  _REGISTERVARIABLE._serialized_start=264
  _REGISTERVARIABLE._serialized_end=350
  _MEMORYVARIABLE._serialized_start=352
  _MEMORYVARIABLE._serialized_end=437
  _STACKVARIABLE._serialized_start=439
  _STACKVARIABLE._serialized_end=556
  _VARIABLEACCESS._serialized_start=559
  _VARIABLEACCESS._serialized_end=843
  _VARIABLEACCESS_VARIABLEACCESSSORT._serialized_start=762
  _VARIABLEACCESS_VARIABLEACCESSSORT._serialized_end=818
  _VARIABLETYPE._serialized_start=845
  _VARIABLETYPE._serialized_end=892
  _VAR2UNIFIED._serialized_start=894
  _VAR2UNIFIED._serialized_end=953
  _PHI2VAR._serialized_start=955
  _PHI2VAR._serialized_end=1002
  _VARIABLEMANAGERINTERNAL._serialized_start=1005
  _VARIABLEMANAGERINTERNAL._serialized_end=1619
# @@protoc_insertion_point(module_scope)



# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("fhsvcctl.dll")
prototypes = \
    {
        # 
        'FhServiceOpenPipe': SimTypeFunction([SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["StartServiceIfStopped", "Pipe"]),
        # 
        'FhServiceClosePipe': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["Pipe"]),
        # 
        'FhServiceStartBackup': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["Pipe", "LowPriorityIo"]),
        # 
        'FhServiceStopBackup': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["Pipe", "StopTracking"]),
        # 
        'FhServiceReloadConfiguration': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["Pipe"]),
        # 
        'FhServiceBlockBackup': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["Pipe"]),
        # 
        'FhServiceUnblockBackup': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["Pipe"]),
    }

lib.set_prototypes(prototypes)



import angr

class sigaction(angr.SimProcedure):
    def run(self, signum, act, oldact): #pylint:disable=arguments-differ,unused-argument
        # TODO: actually do something
        return self.state.solver.BVV(0, self.arch.sizeof['int'])

class rt_sigaction(angr.SimProcedure):
    def run(self, signum, act, oldact, sigsetsize): #pylint:disable=arguments-differ,unused-argument
        # TODO: actually do something
        # ...hack
        if self.state.solver.is_true(signum == 33):
            return self.state.libc.ret_errno('EINVAL')
        return self.state.solver.BVV(0, self.arch.sizeof['int'])



import logging

from ..java import JavaSimProcedure
from angr.engines.soot.values.strref import SimSootValue_StringRef

l = logging.getLogger(name=__name__)


class GetSimpleName(JavaSimProcedure):

    __provides__ = (
        ("java.lang.Class", "getSimpleName()"),
    )

    def run(self, this): # pylint: disable=arguments-differ
        class_simple_name = this.type.split('.')[-1]
        str_ref = SimSootValue_StringRef.new_string(self.state, self.state.solver.StringV(class_simple_name))
        return str_ref



# pylint:disable=unused-argument,useless-return
import ailment

from ...errors import UnsupportedNodeTypeError
from .region_identifier import MultiNode
from .structurer_nodes import CodeNode, SequenceNode, ConditionNode, SwitchCaseNode, LoopNode, CascadingConditionNode, \
    ConditionalBreakNode


class SequenceWalker:
    """
    Walks a SequenceNode and all its nodes, recursively.
    """
    def __init__(self, handlers=None, exception_on_unsupported=False, update_seqnode_in_place=True):
        self._update_seqnode_in_place = update_seqnode_in_place
        self._exception_on_unsupported = exception_on_unsupported

        default_handlers = {
            # Structurer nodes
            CodeNode: self._handle_Code,
            SequenceNode: self._handle_Sequence,
            ConditionNode: self._handle_Condition,
            CascadingConditionNode: self._handle_CascadingCondition,
            SwitchCaseNode: self._handle_SwitchCase,
            LoopNode: self._handle_Loop,
            MultiNode: self._handle_MultiNode,
            ConditionalBreakNode: self._handle_ConditionalBreak,
            ailment.Block: self._handle_Noop,
        }

        self._handlers = default_handlers
        if handlers:
            self._handlers.update(handlers)

    def walk(self, sequence):
        return self._handle(sequence)

    #
    # Handlers
    #

    def _handle(self, node, **kwargs):
        handler = self._handlers.get(node.__class__, None)
        if handler is not None:
            return handler(node, **kwargs)
        if self._exception_on_unsupported:
            raise UnsupportedNodeTypeError("Node type %s is not supported yet." % type(node))
        return None

    def _handle_Code(self, node: CodeNode, **kwargs):
        new_inner_node = self._handle(node.node, parent=node, index=0)
        if new_inner_node is None:
            return None
        return CodeNode(new_inner_node, node.reaching_condition)

    def _handle_Sequence(self, node, **kwargs):
        i = 0
        nodes_copy = list(node.nodes)
        changed = False
        while i < len(nodes_copy):
            node_ = nodes_copy[i]
            new_node = self._handle(node_, parent=node, index=i)
            if new_node is not None:
                changed = True
                if self._update_seqnode_in_place:
                    node.nodes[i] = new_node
                else:
                    nodes_copy[i] = new_node
            i += 1

        if not changed:
            return None
        if self._update_seqnode_in_place:
            return node
        return SequenceNode(node.addr, nodes=nodes_copy)

    def _handle_MultiNode(self, node, **kwargs):
        i = 0
        changed = False
        nodes_copy = list(node.nodes)
        while i < len(nodes_copy):
            node_ = nodes_copy[i]
            new_node = self._handle(node_, parent=node, index=i)
            if new_node is not None:
                changed = True
                node.nodes[i] = new_node
            i += 1
        return None if not changed else node

    def _handle_SwitchCase(self, node, **kwargs):
        self._handle(node.switch_expr, parent=node, label='switch_expr')

        changed = False
        new_cases = { }
        for idx in list(node.cases.keys()):
            case = node.cases[idx]
            new_case = self._handle(case, parent=node, index=idx, label='case')
            if new_case is not None:
                changed = True
                new_cases[idx] = new_case
            else:
                new_cases[idx] = case

        new_default_node = None
        if node.default_node is not None:
            new_default_node = self._handle(node.default_node, parent=node, label='default')
            if new_default_node is not None:
                changed = True
            else:
                new_default_node = node.default_node

        if changed:
            return SwitchCaseNode(node.switch_expr, new_cases, new_default_node, addr=node.addr)

        return None

    def _handle_Loop(self, node: LoopNode, **kwargs):
        if node.initializer is not None:
            self._handle(node.initializer)
        if node.iterator is not None:
            self._handle(node.iterator)
        if node.condition is not None:
            self._handle(node.condition, parent=node, label="condition")
        seq_node = self._handle(node.sequence_node, **kwargs)
        if seq_node is not None:
            return LoopNode(node.sort, node.condition, seq_node, addr=node.addr, continue_addr=node.continue_addr,
                            initializer=node.initializer, iterator=node.iterator)
        return None

    def _handle_Condition(self, node, **kwargs):
        if node.true_node is not None:
            new_true_node = self._handle(node.true_node, parent=node, index=0)
        else:
            new_true_node = None

        if node.false_node is not None:
            new_false_node = self._handle(node.false_node, parent=node, index=1)
        else:
            new_false_node = None

        if new_true_node is None and new_false_node is None:
            return None

        return ConditionNode(node.addr, node.reaching_condition, node.condition,
                             node.true_node if new_true_node is None else new_true_node,
                             false_node=node.false_node if new_false_node is None else new_false_node)

    def _handle_CascadingCondition(self, node: CascadingConditionNode, **kwargs):
        for index, (_, child_node) in enumerate(node.condition_and_nodes):
            self._handle(child_node, parent=node, index=index)
        if node.else_node is not None:
            self._handle(node.else_node, parent=node, index=-1)
        return None

    def _handle_ConditionalBreak(self, node: ConditionalBreakNode, **kwargs):  # pylint:disable=no-self-use
        return None

    def _handle_Noop(self, *args, **kwargs):  # pylint:disable=no-self-use
        return None
